#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2020  Wellcome Trust Sanger Institute and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Generic imports
import os
import sys
import argparse
import re
import math

# Biopython imports
from Bio import AlignIO
from Bio import Phylo
from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq

# command line parsing
def get_options():

    parser = argparse.ArgumentParser(description='Extract all the unique alleles at recombinant loci')

    # input options
    parser.add_argument('--aln',
                        help = 'Input alignment (FASTA format)',
                        required = True)
    parser.add_argument('--gff',
                        help = 'GFF of recombinant regions detected by Gubbins',
                        required = True)
    parser.add_argument('--out-dir',
                        help = 'Output directory',
                        required = True)
    parser.add_argument('--start',
                        help = 'Start of region of interest',
                        default = 1,
                        required = False)
    parser.add_argument('--end',
                        help = 'End of region of interest',
                        default = math.inf,
                        required = False)
    parser.add_argument('--terminal-only',
                        help = 'Only extract recombinations on terminal branches',
                        default = False,
                        action = 'store_true')

    return parser.parse_args()

# main code
if __name__ == "__main__":

    # Get command line options
    args = get_options()

    # Create output directory
    if not os.path.isdir(args.out_dir):
      os.mkdir(args.out_dir)

    # Read recombinant regions from GFF
    rec_start = []
    rec_end = []
    with open(args.gff,'r') as gff_file:
      for line in gff_file.readlines():
        if not line.startswith('##'):
          # Get coordinates
          info = line.rstrip().split('\t')
          taxon_pattern = re.compile('taxa="([^"]*)"')
          taxon_set = set(taxon_pattern.search(info[8]).group(1).split())
          if (len(taxon_set) == 1 or not args.terminal_only):
            rec_start.append(int(info[3]))
            rec_end.append(int(info[4]))
    
    # Read in alignment and identify recombinations
    alignment = AlignIO.read(args.aln,'fasta')
    for (start,end) in zip(rec_start,rec_end):
      if start >= args.start and end <= args.end:
        out_fn = 'locus_' + str(start) + '_' + str(end) + '.aln'
        with open(os.path.join(args.out_dir,out_fn),'w') as out_file:
          seen_seqs = []
          rec_locus_alignment = alignment[:,(start-1):end]
          for taxon in rec_locus_alignment:
            if taxon.seq not in seen_seqs:
              out_file.write('>' + taxon.id + '\n' + str(taxon.seq) + '\n')
              seen_seqs.append(taxon.seq)



#!/usr/bin/env python
# encoding: utf-8
#
# Wellcome Trust Sanger Institute
# Copyright (C) 2012  Wellcome Trust Sanger Institute
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import sys
import argparse
from gubbins.__init__ import version
import gubbins.common

def parse_input_args():

    parser = argparse.ArgumentParser(
        description='Croucher N. J., Page A. J., Connor T. R., Delaney A. J., Keane J. A., Bentley S. D., Parkhill J., '
                    'Harris S.R. "Rapid phylogenetic analysis of large samples of recombinant bacterial whole genome '
                    'sequences using Gubbins". Nucleic Acids Res. 2015 Feb 18;43(3):e15. doi: 10.1093/nar/gku1196.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    ioGroup = parser.add_argument_group('Input and output options')
    ioGroup.add_argument('alignment_filename',        help='Multifasta alignment file')
    ioGroup.add_argument('--prefix',            '-p', help='Add a prefix to the final output filenames')
    ioGroup.add_argument('--starting-tree',     '-s', help='Starting tree')
    ioGroup.add_argument('--date',              '-D', help='Two-column text file in which the second column is the'
                                                      ' date of isolation in YYYY,YYYY-MM or YYYY-MM-DD format',
                                                      default = None)
    ioGroup.add_argument('--use-time-stamp',    '-u', help='Use a time stamp in file names', action='store_true')
    ioGroup.add_argument('--version',                 action='version',
                                                      version = version())
    ioGroup.add_argument('--threads',           '-c', help='Number of threads to use for parallelisation',
                                                      type=int,  default=1)
    ioGroup.add_argument('--verbose',           '-v', help='Turn on debugging', action='store_true')
    ioGroup.add_argument('--no-cleanup',        '-n', help='Do not cleanup intermediate files', action='store_true')

    dataGroup = parser.add_argument_group('Data processing options')
    dataGroup.add_argument('--pairwise',              help='Compare two sequences (without using a tree)',
                                                      default = False, action = 'store_true') # fasttree model fit, star phylogeny, one iteration
    dataGroup.add_argument('--filter-percentage','-f',
                                                      help='Filter out taxa with more than this percentage of gaps',
                                                      type=float, default=25.0)
    dataGroup.add_argument('--remove-identical-sequences',
                                                '-d', help='Remove identical sequences', action='store_true')

    treeGroup = parser.add_argument_group('Tree building options')
    treeGroup.add_argument('--tree-builder',    '-t', help='Application to use for tree building',
                                                      default='raxml',
                                                      choices=['raxml', 'raxmlng', 'iqtree', 'iqtree-fast', 'fasttree', 'hybrid', 'rapidnj','veryfasttree'])
    treeGroup.add_argument('--tree-args',             help='Quoted string of further arguments passed to tree building algorithm'
                                                      ' (start string with a space if there is a risk of being interpreted as a flag)',
                                                      default = None)
    treeGroup.add_argument('--first-tree-builder',    help='Application to use for building the first tree',
                                                      default=None,
                                                      choices=['raxml', 'raxmlng', 'iqtree', 'iqtree-fast', 'fasttree', 'rapidnj', 'star','veryfasttree'])
    treeGroup.add_argument('--first-tree-args',       help='Further arguments passed to first tree building algorithm',
                                                      default = None)
    treeGroup.add_argument('--outgroup',        '-o', help='Outgroup name for rerooting. A list of comma separated '
                                                      'names can be used if they form a clade',
                                                      default = None)
    treeGroup.add_argument('--bootstrap',       '-#', help='Number of bootstrap replicates to perform with final alignment',
                                                      type = int, default = 0)
    treeGroup.add_argument('--transfer-bootstrap',    help='Calculate bootstrap supporting transfer bootstrap expectation',
                                                      default = False, action = 'store_true')
    treeGroup.add_argument('--sh-test',               help='Perform an SH test of node likelihoods', default = False,
                                                      action = 'store_true')
    treeGroup.add_argument('--seed',                  help='Set seed for reproducibility of analysis',
                                                      default = None, type = int)
    treeGroup.add_argument('--invariant-site-correction',
                                                      help='Correct for invariant sites',
                                                      default = False, action = 'store_true')
                                                          
    modelGroup = parser.add_argument_group('Nucleotide substitution model options')
    modelGroup.add_argument('--model',          '-M', help='Nucleotide substitution model (not all available for all '
                                                      'tree building algorithms)',
                                                      default=None,
                                                      choices=['JC','K2P','HKY','GTR','GTRGAMMA','GTRCAT'])
    modelGroup.add_argument('--first-model',          help='Nucleotide substitution model used for first tree',
                                                      default=None,
                                                      choices=['JC','K2P','HKY','GTR','GTRGAMMA','GTRCAT'])
    modelGroup.add_argument('--best-model',           help='Automatically select best substitution model using iqtree in later iterations',
                                                      default = False, action = 'store_true')
    modelGroup.add_argument('--custom-model',         help='String corresponding to a substitution model for the selected tree'
                                                      ' building algorithm', default = None)
    modelGroup.add_argument('--custom-first-model',   help='String corresponding to a substitution model for the selected tree'
                                                      ' building algorithm for the first iteration',
                                                      default = None)

    reconGroup = parser.add_argument_group('Ancestral sequence reconstruction options')
    reconGroup.add_argument('--model-fitter',   '-F', help='Application to use for model fitting for joint ancestral state'
                                                      ' reconstruction [if unspecified: same as tree builder if possible'
                                                      ', else iqtree]',
                                                      default = None,
                                                      choices=['raxml', 'raxmlng', 'iqtree', 'fasttree', None])
    reconGroup.add_argument('--recon-model',    '-R', help='Nucleotide substitution model used for ancestral state reconstruction'
                                                      ' (not all available for all tree building algorithms)',
                                                      default='GTRGAMMA',
                                                      choices=['JC','K2P','HKY','GTR','GTRGAMMA','GTRCAT'])
    reconGroup.add_argument('--custom-recon-model',   help='String corresponding to a substitution model for the selected '
                                                      ' model fitting algorithm',
                                                      default=None)
    reconGroup.add_argument('--recon-with-dates',     help='Use isolate date information in ancestral joint sequence'
                                                      ' reconstruction',
                                                      default=False, action='store_true')
    reconGroup.add_argument('--model-fitter-args',    help='Further arguments passed to model fitting algorithm',
                                                      default=None)
    reconGroup.add_argument('--mar',                  help='Use marginal, rather than joint, ancestral reconstruction',
                                                      action='store_true')
    reconGroup.add_argument('--seq-recon',            help='Algorithm to use for marginal reconstruction [if unspecified: '
                                                      'same as tree builder if possible, else iqtree; requires --mar flag]',
                                                      default=None,
                                                      choices=['raxml', 'raxmlng', 'iqtree', None])
    reconGroup.add_argument('--seq-recon-args',       help='Further arguments passed to sequence reconstruction algorithm'
                                                      ' (start string with a space if there is a risk of being interpreted as a flag)',
                                                      default=None)
                                                            
    gubbinsGroup = parser.add_argument_group('Recombination detection options')
    gubbinsGroup.add_argument('--min-snps',     '-m', help='Min SNPs to identify a recombination block',
                                                      type=int,
                                                      default = 3)
    gubbinsGroup.add_argument('--min-window-size','-a',
                                                      help='Minimum window size', type=int, default=100)
    gubbinsGroup.add_argument('--max-window-size','-b',
                                                      help='Maximum window size', type=int, default=10000)
    gubbinsGroup.add_argument('--p-value',
                                                      help='Uncorrected p value used to identify recombinations',
                                                      type=float,
                                                      default=0.05)
    gubbinsGroup.add_argument('--trimming-ratio',
                                                      help='Ratio of log probabilities used to trim recombinations',
                                                      type=float,
                                                      default=1.0)
    gubbinsGroup.add_argument('--extensive-search',
                                                      help='Undertake slower, more thorough, search for recombination',
                                                      action='store_true',
                                                      default=False)

    stopGroup = parser.add_argument_group('Algorithm start/stop options')
    stopGroup.add_argument('--iterations',      '-i',
                                                        help='Maximum No. of iterations',
                                                        type=int,
                                                        default=5)
    stopGroup.add_argument('--converge-method', '-z',
                                                        help='Criteria to use to know when to halt iterations',
                                                        default='weighted_robinson_foulds',
                                                        choices=['weighted_robinson_foulds', 'robinson_foulds',
                                                                     'recombination'])
    stopGroup.add_argument('--resume',
                                                        help='Intermediate tree from previous run (must include'
                                                        ' "iteration_X" in file name)',
                                                        default=None)
    return parser


def main():
    parser = parse_input_args()
    gubbins.common.parse_and_run(parser.parse_args(), parser.description)

if __name__ == '__main__':
    main()
    sys.exit(0)



#! /usr/bin/env python3
# encoding: utf-8

"""
Test manipulating and interogating trees using python modules with no external application dependancies.
"""

import unittest
import shutil
import os
import filecmp
from gubbins import common, treebuilders

modules_dir = os.path.dirname(os.path.abspath(treebuilders.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')

class TestTreeMethods(unittest.TestCase):

    def initialise_Star(self):
        algorithm = treebuilders.Star()
        assert algorithm.name == 'Star'

    def initialise_RapidNJ(self):
        algorithm = treebuilders.RapidNJ()
        assert algorithm.name == 'RapidNJ'

    def initialise_FastTree(self):
        algorithm = treebuilders.FastTree()
        assert algorithm.name == 'FastTree'

    def initialise_IQTree(self):
        algorithm = treebuilders.IQTree()
        assert algorithm.name == 'IQTree'

    def initialise_RAxML(self):
        algorithm = treebuilders.RAxML()
        assert algorithm.name == 'RAxML'

    def initialise_RAxMLNG(self):
        algorithm = treebuilders.RAxMLNG()
        assert algorithm.name == 'RAxMLNG'

    def initialise_VeryFastTree(self):
        algorithm = treebuilders.VeryFastTree()
        assert algorithm.name == 'VeryFastTree'

    def test_robinson_foulds_distance(self):
        # two tree with different distances
        assert common.robinson_foulds_distance(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                                               os.path.join(data_dir, 'robinson_foulds_distance_tree2.tre')) == 17.263494
        # two trees with same distance
        assert common.robinson_foulds_distance(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                                               os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre')) == 0

    def test_has_tree_been_seen_before(self):
        # same content so the tree has been seen before
        assert common.has_tree_been_seen_before([os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                                                 os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                                                 os.path.join(data_dir, 'robinson_foulds_distance_tree1_dup.tre')],
                                                'weighted_robinson_foulds') == 1
        # different trees
        assert common.has_tree_been_seen_before([os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                                                 os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                                                 os.path.join(data_dir, 'robinson_foulds_distance_tree2.tre')],
                                                'weighted_robinson_foulds') == 0

    def test_root_tree(self):
        common.root_tree(os.path.join(data_dir, 'unrooted_tree.newick'), os.path.join(data_dir, 'actual_rooted_tree.newick'))
        assert filecmp.cmp(os.path.join(data_dir, 'actual_rooted_tree.newick'),
                           os.path.join(data_dir, 'expected_rooted_tree.newick'), shallow=False)
        os.remove(os.path.join(data_dir, 'actual_rooted_tree.newick'))

    def test_reroot_tree(self):
        shutil.copyfile(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                        os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_test'))
        common.reroot_tree(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_test'),
                           'sequence_4')
        assert filecmp.cmp(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_test'),
                           os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_expected'))
        os.remove(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_test'))

        shutil.copyfile(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                        os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_test'))
        common.reroot_tree(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_test'), '')
        assert filecmp.cmp(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_test'),
                           os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_expected'))
        os.remove(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_test'))

    def test_reroot_tree_with_outgroup(self):
        shutil.copyfile(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                        os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_list_test'))
        common.reroot_tree_with_outgroup(
            os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_list_test'), ['sequence_4'])
        assert filecmp.cmp(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_list_test'),
                           os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_expected'))
        os.remove(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_list_test'))

    def test_reroot_tree_with_outgroups(self):
        shutil.copyfile(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                        os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_2_test'))
        common.reroot_tree_with_outgroup(
            os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_2_test'),
            ['sequence_4', 'sequence_2'])
        assert filecmp.cmp(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_2_test'),
                           os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_2_expected'))
        os.remove(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_at_sequence_4_2_test'))

    def test_reroot_tree_with_outgroups_all_in_one_clade(self):
        outgroups = ['A', 'B']
        expected_monophyletic_outgroup = ['A', 'B']
        expected_output_file = os.path.join(data_dir, 'expected_reroot_tree_with_outgroups_all_in_one_clade.tre')
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def test_reroot_tree_with_outgroups_all_in_one_clade_large(self):
        outgroups = ['A', 'B', 'C']
        expected_monophyletic_outgroup = ['A', 'B', 'C']
        expected_output_file = os.path.join(data_dir, 'expected_test_reroot_tree_with_outgroups_all_in_one_clade_large.tre')
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def test_reroot_tree_with_outgroups_all_in_different_clade(self):
        outgroups = ['A', 'D']
        expected_monophyletic_outgroup = ['A']
        expected_output_file = os.path.join(data_dir, 'expected_reroot_tree_with_outgroups_all_in_different_clade.tre')
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def test_reroot_tree_with_outgroups_with_two_mixed_clades(self):
        outgroups = ['A', 'B', 'C', 'D']
        expected_monophyletic_outgroup = ['A']
        expected_output_file = os.path.join(data_dir, 'expected_reroot_tree_with_outgroups_with_two_mixed_clades.tre')
        self.reroot_tree_check(outgroups, expected_output_file, expected_monophyletic_outgroup)

    def reroot_tree_check(self, outgroups, expected_output_file, expected_monophyletic_outgroup):
        tmp_tree_file_name = expected_output_file.replace('expected_','generated_')
        shutil.copyfile(os.path.join(data_dir, 'outgroups_input.tre'), tmp_tree_file_name)
        assert expected_monophyletic_outgroup == common.get_monophyletic_outgroup(tmp_tree_file_name, outgroups)
        common.reroot_tree_with_outgroup(tmp_tree_file_name, outgroups)
        assert filecmp.cmp(tmp_tree_file_name, expected_output_file)
        os.remove(tmp_tree_file_name)

    def test_split_all_non_bi_nodes(self):
        # best way to access it is via reroot_tree_at_midpoint because it outputs to a file
        shutil.copyfile(os.path.join(data_dir, 'non_bi_tree.tre'), os.path.join(data_dir, 'non_bi_tree.tre.actual'))
        common.reroot_tree_at_midpoint(os.path.join(data_dir, 'non_bi_tree.tre.actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'non_bi_tree.tre.actual'), os.path.join(data_dir, 'non_bi_tree.tre.expected'))
        os.remove(os.path.join(data_dir, 'non_bi_tree.tre.actual'))

    def test_reroot_tree_at_midpoint(self):
        shutil.copyfile(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
                        os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual'))
        common.reroot_tree_at_midpoint(
            os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual'))
        assert filecmp.cmp(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual'),
                           os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_expected'))
        os.remove(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.reroot_tree_at_midpoint_actual'))

    def test_filter_out_removed_taxa_from_tree(self):
        common.filter_out_removed_taxa_from_tree(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre'),
             os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_actual'),
             ['sequence_1', 'sequence_2', 'sequence_3', 'sequence_4', 'sequence_5'])
        assert filecmp.cmp(
            os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_actual'),
            os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_expected'))
        os.remove(os.path.join(data_dir, 'robinson_foulds_distance_tree1.tre.filter_out_removed_taxa_from_tree_actual'))

    def test_internal_node_taxons_removed_when_used_as_starting_tree(self):
        common.filter_out_removed_taxa_from_tree(os.path.join(data_dir, 'tree_with_internal_nodes.tre'),
                                                 os.path.join(data_dir, 'tree_with_internal_nodes.tre_actual'), [])
        assert filecmp.cmp(os.path.join(data_dir, 'tree_with_internal_nodes.tre_actual'),
                           os.path.join(data_dir, 'tree_with_internal_nodes.tre_expected'))
        os.remove(os.path.join(data_dir, 'tree_with_internal_nodes.tre_actual'))

    def test_transfer_internal_node_labels_to_tree(self):
        reconstructor = treebuilders.IQTree(1, model = 'GTRGAMMA', internal_node_prefix = 'internal_')
        common.transfer_internal_node_labels_to_tree(os.path.join(data_dir, 'source_tree.tre'),
                                                     os.path.join(data_dir, 'destination_tree.tre'),
                                                     os.path.join(data_dir, 'renamed_output_tree'), reconstructor)
        assert os.path.exists(os.path.join(data_dir, 'renamed_output_tree'))
        assert filecmp.cmp(os.path.join(data_dir, 'renamed_output_tree'),
                           os.path.join(data_dir, 'expected_renamed_output_tree'), shallow=False)
        os.remove(os.path.join(data_dir, 'renamed_output_tree'))

    def test_remove_internal_node_labels(self):
        common.remove_internal_node_labels_from_tree(os.path.join(data_dir, 'final_tree_with_internal_labels.tre'),
                                                     'final_tree_with_internal_labels.tre')
        assert os.path.exists('final_tree_with_internal_labels.tre')
        assert filecmp.cmp('final_tree_with_internal_labels.tre',
                           os.path.join(data_dir, 'expected_final_tree_without_internal_labels.tre'))
        os.remove('final_tree_with_internal_labels.tre')


if __name__ == "__main__":
    unittest.main()



#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2020  Wellcome Trust Sanger Institute and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Generic imports
import sys
import argparse
import re
# Phylogenetic imports
import dendropy
# Biopython imports
from Bio import AlignIO
from Bio import Phylo
from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq

# command line parsing
def get_options():

    parser = argparse.ArgumentParser(description='Extract a clade from a Gubbins output',
                                     prog='extract_gubbins_clade')

    # input options
    parser.add_argument('--clades',
                        help = 'Two column file assigning isolates (first column) to clades (second column)',
                        required = True)
    parser.add_argument('--gff',
                        help = 'recombination prediction GFF file output by Gubbins',
                        required = True)
    parser.add_argument('--snps',
                        help = 'branch base reconstruction EMBL file output by Gubbins',
                        required = True)
    parser.add_argument('--exclude-regions',
                        help = 'Two column file specifying start and end of regions to be excluded',
                        required = False,
                        default = None)
    parser.add_argument('--tree',
                        help = 'Labelled tree output by Gubbins',
                        required = True)
    parser.add_argument('--print-trees',
                        help = 'Print clade trees',
                        default = False,
                        action = 'store_true')
    parser.add_argument('--print-rec-lengths',
                        help = 'Print recombination lengths',
                        default = False,
                        action = 'store_true')
    parser.add_argument('--out',
                        help = 'Output file prefix; suffix is "_clades.csv"',
                        required = True)

    return parser.parse_args()

# main code
if __name__ == "__main__":

    # Get command line options
    args = get_options()
    
    # Parse clades
    clades = {}
    clade_names = set()
    with open(args.clades,'r') as clade_list:
        for line in clade_list.readlines():
            info = line.strip().split()
            if len(info) == 2:
                clades[info[0]] = info[1]
                clade_names.add(info[1])
            else:
                sys.stderr.write('Line needs two columns: ' + line + '\n')
    
    # Exclude regions
    excluded_region_starts = []
    excluded_region_ends = []
    if args.exclude_regions is not None:
        with open(args.exclude_regions,'r') as exclude_file:
            for line in exclude_file.readlines():
                coords = line.strip().split()
                if int(coords[0]) < int(coords[1]):
                    excluded_region_starts.append(int(coords[0]))
                    excluded_region_ends.append(int(coords[1]))
                else:
                    sys.stderr.write('Start of excluded region must be less than end\n')
                    sys.exit(1)
    
    # Store SNP information
    node_snps = {}
    snp_total = 0
    with open(args.snps,'r') as snp_file:
        pos = 0
        for line in snp_file.readlines():
            info = line.strip().split()
            if info[1] == 'variation':
                pos = int(info[2])
            if info[1].startswith('/node='):
                node = info[1].replace('"','').split('->')
                include_snp = True
                for s,e in zip(excluded_region_starts,excluded_region_ends):
                    if pos >= s and pos <= e:
                        include_snp = False
                        break
                if include_snp:
                    snp_total += 1
                    if node[1] in node_snps:
                        node_snps[node[1]].append(pos)
                    else:
                        node_snps[node[1]] = [pos]

    # Store recombination information
    node_rec_starts = {}
    node_rec_ends = {}
    with open(args.gff,'r') as gff_file:
        for line in gff_file.readlines():
            if not line.startswith('##'):
                info = line.rstrip().split('\t')
                start = int(info[3])
                end = int(info[4])
                node = info[8].split(';')[0].replace('"','').split('->')[1]
                include_rec = True
                for s,e in zip(excluded_region_starts,excluded_region_ends):
                    if start >= s and end <= e:
                        include_rec = False
                if include_rec:
                    if node not in node_rec_starts:
                        node_rec_starts[node] = [start]
                        node_rec_ends[node] = [end]
                    else:
                        node_rec_starts[node].append(start)
                        node_rec_ends[node].append(end)
    
    # Divide SNPs into recombinant and non-recombinant
    rec_snps = {node:0 for node in node_snps}
    pm_snps = {node:0 for node in node_snps}
    for node in node_snps:
        for p in node_snps[node]:
            rec_snp = False
            if node in node_rec_starts:
                for s,e in zip(node_rec_starts[node],node_rec_ends[node]):
                    if p >= s and p <= e:
                        rec_snp = True
                        break
            if rec_snp:
                rec_snps[node] += 1
            else:
                pm_snps[node] += 1
    
    # Parse tree
    info_labels = ['total_snps','rec_snps','mutation_snps','recombinations']
    tree_info_labels = ['n_taxa','n_branches','branch_length']
    tree = dendropy.Tree.get(path = args.tree,
                             schema = 'newick',
                             preserve_underscores = True,
                             rooting='force-rooted')
    taxon_names = [taxon.label for taxon in tree.taxon_namespace]
    
    # Calculate statistics per clade
    rec_length_string = ''
    with open(args.out + '_clades.csv','w') as out_file:
        out_file.write('Clade,')
        out_file.write(','.join(info_labels + tree_info_labels))
        out_file.write('\n')
        for clade_name in clade_names:
            out_file.write(clade_name + ',')
            clade_members = [sequence for sequence in clades if clades[sequence] == clade_name]
            # Check all isolates are in the tree
            for isolate in clade_members:
                if isolate not in taxon_names:
                    sys.stderr.write('Isolate ' + isolate + ' not in tree\n')
                    sys.stderr.write('Tree contains ' + ','.join(taxon_names) + '\n')
                    #exit(1)
            # Extract clade tree
            clade_tree = tree.clone(depth = 1)
            clade_tree.retain_taxa_with_labels(clade_members)
            # Print tree
            if args.print_trees:
                clade_tree_string = clade_tree.as_string(
                                        schema='newick',
                                        suppress_leaf_taxon_labels=False,
                                        suppress_leaf_node_labels=True,
                                        suppress_internal_taxon_labels=True,
                                        suppress_internal_node_labels=True,
                                        suppress_rooting=True,
                                        suppress_edge_lengths=False,
                                        unquoted_underscores=True,
                                        preserve_spaces=False,
                                        store_tree_weights=False,
                                        suppress_annotations=True,
                                        annotations_as_nhx=False,
                                        suppress_item_comments=True,
                                        node_label_element_separator=' '
                                    )
                with open(clade_name + '.tre','w') as tree_out:
                    tree_out.write(clade_tree_string.replace('\'', '') + '\n')
            # Print statistics
            clade_info = {label:0 for label in info_labels + tree_info_labels}
            for node in clade_tree.preorder_node_iter():
                if node != clade_tree.seed_node:
                    clade_info['n_branches'] += 1
                    clade_info['branch_length'] += node.edge_length
                    if node.is_leaf():
                        clade_info['n_taxa'] += 1
                        node_label_string = node.taxon.label
                    else:
                        node_label_string = node.label
                    if node_label_string in node_snps:
                        clade_info['total_snps'] += len(node_snps[node_label_string])
                        clade_info['rec_snps'] += rec_snps[node_label_string]
                        clade_info['mutation_snps'] += pm_snps[node_label_string]
                    if node_label_string in node_rec_starts:
                        clade_info['recombinations'] += len(node_rec_starts[node_label_string])
                        if args.print_rec_lengths:
                            for s,e in zip(node_rec_starts[node_label_string],node_rec_ends[node_label_string]):
                                rec_length_string += clade_name + ',' + str(1+e-s) + '\n'
                    
            out_file.write(','.join([str(clade_info[label]) for label in info_labels + tree_info_labels]))
            out_file.write('\n')
    
    if args.print_rec_lengths:
        with open(args.out + '_rec_lengths.csv','w') as rec_out_file:
            rec_out_file.write('Clade,Length\n' + rec_length_string)



#!/usr/bin/env python
# encoding: utf-8
#
# pyjar written by Simon Harris
# code modified from https://github.com/simonrharris/pyjar
# pyjar is free software, licensed under GPLv3.

from asyncio import subprocess
import shutil
from scipy import linalg
import numpy
import dendropy
import sys
import os
import time
from Bio import AlignIO
from math import log, exp
from functools import partial
import numba
from numba import jit, njit, types, from_dtype
from math import ceil
import collections
import datetime
import multiprocessing
try:
    from multiprocessing import Pool, shared_memory
    from multiprocessing.managers import SharedMemoryManager
    NumpyShared = collections.namedtuple('NumpyShared', ('name', 'shape', 'dtype'))
except ImportError as e:
    sys.stderr.write("This version of Gubbins requires the multiprocessing library and python v3.8 or higher for memory management\n")
    sys.exit(201)

from gubbins.utils import generate_shared_mem_array

###########################
# Python-native functions #
###########################

class Pyjar:
    def __init__(self, model: str):
        """Initialises the object"""
        self.citation = "https://doi.org/10.1093/oxfordjournals.molbev.a026369"
        self.process = "Sequence reconstructor"
        self.version = "1.0"
        self.algorithm = "pyjar"
        self.executable = "pyjar"
        self.model = model

# Split a list into chunks for multiprocessing
# from https://stackoverflow.com/questions/2130016/splitting-a-list-into-n-parts-of-approximately-equal-length/37414115#37414115
def chunks(l, k):
    n = len(l)
    return [l[i * (n // k) + min(i, n % k):(i+1) * (n // k) + min(i+1, n % k)] for i in range(k)]

#Read the tree file and root
def read_tree(treefile):
    if not os.path.isfile(treefile):
        print("Error: tree file does not exist")
        sys.exit(204)
    t=dendropy.Tree.get(path=treefile,
                        schema="newick",
                        preserve_underscores=True,
                        rooting="force-rooted")
    return t

#Calculate Pij from Q matrix and branch length
def calculate_pij(branch_length,rate_matrix):
    if branch_length==0:
        pij = numpy.full((4,4), -numpy.inf, dtype = numpy.float32)
        numpy.fill_diagonal(pij, 0.0)
    else:
        pij = numpy.array(numpy.log(linalg.expm(numpy.multiply(branch_length,rate_matrix))), dtype = numpy.float32) # modified
    return pij.flatten()

# Create an instanteous rate matrix
def create_rate_matrix(f, r):
    #convert f and r to Q matrix
    rm=numpy.array([[0, f[0]*r[1], f[0]*r[2], f[0]*r[3]],[f[1]*r[0], 0, f[1]*r[3],f[1]*r[4]],[f[2]*r[1], f[2]*r[3], 0, f[2]*r[5]],[f[3]*r[2], f[3]*r[4], f[3]*r[5], 0]])

    rm[0][0]=numpy.sum(rm[0])*-1
    rm[1][1]=numpy.sum(rm[1])*-1
    rm[2][2]=numpy.sum(rm[2])*-1
    rm[3][3]=numpy.sum(rm[3])*-1

    return rm

# Read the info file from the selected phylogenetic software
# to get rates and frequencies
def read_info(infofile, type = 'raxml'):

    if not os.path.isfile(infofile):
        print("Error: model information file " + infofile + " does not exist")
        sys.exit(205)
    
    if type not in ['raxml', 'raxmlng', 'iqtree','fasttree']:
        sys.stderr.write('Only able to parse GTR-type models from raxml, iqtree or fasttree')
        sys.exit(206)
    
    r=[-1.0] * 6 # initialise rates
    f=[-1.0] * 4 # initialise frequencies
    
    with open(infofile, "r") as info_file:
        for line in info_file:
            line=line.strip()
            if type == 'raxml':
                if "freq pi" in line:
                    words=line.split()
                    if "pi(A)" in line:
                        f[0] = float(words[2])
                    elif "pi(C)" in line:
                        f[1] = float(words[2])
                    elif "pi(G)" in line:
                        f[2] = float(words[2])
                    elif "pi(T)" in line:
                        f[3] = float(words[2])
                elif "Base frequencies:" in line:
                    words=line.split()
                    f=[float(words[2]), float(words[3]), float(words[4]), float(words[5])]
                elif "<->" in line:
                    # order is ac ag at cg ct gt
                    words=line.split()
                    if "A <-> C" in line:
                        r[0] = float(words[4])
                    elif "A <-> G" in line:
                        r[1] = float(words[4])
                    elif "A <-> T" in line:
                        r[2] = float(words[4])
                    elif "C <-> G" in line:
                        r[3] = float(words[4])
                    elif "C <-> T" in line:
                        r[4] = float(words[4])
                    elif "G <-> T" in line:
                        r[5] = float(words[4])
                elif "alpha[0]:" in line:
                    # order is ac ag at cg ct gt
                    words=line.split()
                    r=[float(words[9]), float(words[10]), float(words[11]), float(words[12]), float(words[13]), float(words[14])]
            elif type == 'raxmlng':
                sep_by_braces = line.replace('{','}').split('}')
                if sep_by_braces[0] == "GTR":
                    r = [float(rate) for rate in sep_by_braces[1].split('/')]
                    f = [float(rate) for rate in sep_by_braces[3].split('/')]
                elif sep_by_braces[0] == "K80":
                    sep_rates = [float(rate) for rate in sep_by_braces[1].split('/')]
                    r = [sep_rates[0], sep_rates[1], sep_rates[0], sep_rates[0], sep_rates[1], sep_rates[0]]
                    f = [0.25,0.25,0.25,0.25]
                elif sep_by_braces[0] == "HKY":
                    sep_rates = [float(rate) for rate in sep_by_braces[1].split('/')]
                    r = [sep_rates[0], sep_rates[1], sep_rates[0], sep_rates[0], sep_rates[1], sep_rates[0]]
                    f = [float(rate) for rate in sep_by_braces[3].split('/')]
                elif line.startswith("JC"):
                    f = [0.25,0.25,0.25,0.25]
                    r = [1.0,1.0,1.0,1.0,1.0,1.0]
            elif type == 'iqtree':
                # Base frequencies
                words=line.split()
                if "pi(A)" in line:
                    f[0] = float(words[2])
                elif "pi(C)" in line:
                    f[1] = float(words[2])
                elif "pi(G)" in line:
                    f[2] = float(words[2])
                elif "pi(T)" in line:
                    f[3] = float(words[2])
                # Rate estimates
                if "A-C:" in line:
                    r[0] = float(words[1])
                elif "A-G:" in line:
                    r[1] = float(words[1])
                elif "A-T:" in line:
                    r[2] = float(words[1])
                elif "C-G:" in line:
                    r[3] = float(words[1])
                elif "C-T:" in line:
                    r[4] = float(words[1])
                elif "G-T:" in line:
                    r[5] = float(words[1])
            
                if line.startswith('Base frequencies:'):
                    words=line.split()
                    f=[float(words[3]), float(words[5]), float(words[7]), float(words[9])]
                elif line.startswith('Rate parameters:'):
                    words=line.split()
                    # order is ac ag at cg ct gt
                    r=[float(words[3]), float(words[5]), float(words[7]), float(words[9]), float(words[11]), float(words[13])]
            elif type == 'fasttree':
                if line.startswith('GTRFreq'):
                    words=line.split()
                    f=[float(words[1]), float(words[2]), float(words[3]), float(words[4])]
                elif line.startswith('GTRRates'):
                    words=line.split()
                    r=[float(words[1]), float(words[2]), float(words[3]), float(words[4]), float(words[5]), float(words[6])]
                elif 'ML Model: Jukes-Cantor' in line: # Jukes-Cantor model for Fasttree
                    f = [0.25,0.25,0.25,0.25]
                    r = [1.0,1.0,1.0,1.0,1.0,1.0]

    # Check frequencies and rates have been extracted correctly
    if -1.0 in f or -1.0 in r:
        sys.stderr.write('Problem with extracting model parameters - frequencies are ' + str(f) + ' and rates are ' + str(r))
        sys.exit(207)

    return numpy.array(f, dtype = numpy.float32), numpy.array(r, dtype = numpy.float32)

# Convert arrays of variable-length arrays to a square matrix
# for compatibility with numba
# from https://stackoverflow.com/questions/32037893/numpy-fix-array-with-rows-of-different-lengths-by-filling-the-empty-elements-wi
def convert_to_square_numpy_array(data):
    # Get lengths of each row of data
    lens = numpy.array([len(i) for i in data])

    # Mask of valid places in each row
    mask = numpy.arange(lens.max()) < lens[:,None]

    # Setup output array and put elements from data into masked positions
    out = numpy.full(mask.shape, -1, dtype = numpy.int32)
    out[mask] = numpy.concatenate(data)
    return out

# Get the unique base patterns within the numpy array
# Based on https://stackoverflow.com/questions/21888406/getting-the-indexes-to-the-duplicate-columns-of-a-numpy-array
def get_unique_columns(data):
    dt = numpy.dtype((numpy.void, data.dtype.itemsize * data.shape[0]))
    dataf = numpy.asfortranarray(data).view(dt)
    u,uind = numpy.unique(dataf, return_inverse=True)
    u = u.view(data.dtype).reshape(-1,data.shape[0]).T
    return (u,uind)

##########################
# JIT-compiled functions #
##########################

# Convert bases to integers
###########################
@njit(numba.void(numba.typeof(numpy.dtype('U1'))[:],
                numba.uint8[:]),
                cache = True)
def seq_to_int(seq,out_seq):
    for i,b in enumerate(seq):
        if b == 'A':
            out_seq[i] = 0
        elif b == 'C':
            out_seq[i] = 1
        elif b == 'G':
            out_seq[i] = 2
        elif b == 'T':
            out_seq[i] = 3
        elif b == '-':
            out_seq[i] = 4
        elif b == 'N':
            out_seq[i] = 5
        else:
            print('Unable to process character',b)

# Calculate most likely base given bases in descendents
#######################################################
@njit(numba.void(numba.float32[:,:],
                numba.uint8[:,:],
                numba.float32[:,::1],
                numba.int32,
                numba.int32[:,:],
                numba.uint8[::1]),
                cache=True)
def find_most_likely_base_given_descendents(Lmat, Cmat, pij, node_index, child_nodes, column_base_indices):
    #2a. Lz(i) = maxj Pij(tz) x Lx(j) x Ly(j)
    #2b. Cz(i) = the value of j attaining the above maximum.
    child_node_indices = child_nodes[node_index,:]
    child_node_indices = child_node_indices[child_node_indices > -1]
    for end_index in column_base_indices:
        c = Lmat[child_node_indices,end_index].sum()
        for start_index in column_base_indices:
            j = pij[start_index,end_index]+c
            if j > Lmat[node_index,start_index]:
                Lmat[node_index,start_index] = j
                Cmat[node_index,start_index] = end_index

# Calculate the most likely base at the root node
#################################################
@njit(numba.void(numba.float32[:,:],
                numba.uint8[:,:],
                numba.float32[:],
                numba.int32,
                numba.int32[:],
                numba.uint8[:]),
                cache=True)
def calculate_root_likelihood(Lmat, Cmat, base_frequencies, node_index, child_node_indices, column_base_indices):
    for end_index in column_base_indices:
        c = Lmat[child_node_indices,end_index].sum()
        for start_index in column_base_indices:
            j = log(base_frequencies[end_index]) + c
            if j > Lmat[node_index,start_index]:
                Lmat[node_index,start_index] = j
                Cmat[node_index,start_index] = end_index

# Fill in matrices given known or unknown base in sequence
##########################################################
@njit(numba.void(numba.float32[:,:],
                numba.uint8[:,:],
                numba.float32[:,:],
                numba.int32,
                numba.uint8),
                cache=True)
def process_leaf(Lmat, Cmat, pij, node_index, taxon_base_index):
    if taxon_base_index < 4:
        #1a. Let j be the amino acid at y. Set, for each amino acid i: Cy(i)= j. This implies that no matter what is the amino acid in the father of y, j is assigned to node y.
        Cmat[node_index,:] = taxon_base_index
        #1b. Set for each amino acid i: Ly(i) = Pij(ty), where ty is the branch length between y and its father.
        Lmat[node_index,:] = [pij[0][taxon_base_index],pij[1][taxon_base_index],pij[2][taxon_base_index],pij[3][taxon_base_index]]
    else:
        # Cmat stays as default when base is unknown
        Lmat[node_index,:] = [pij[0][0],pij[1][1],pij[2][2],pij[3][3]]

# Count the number of substitutions occurring on a branch
#########################################################
@njit(numba.void(numba.int32[:],
                numba.int32[:],
                numba.int32[:],
                numba.int32,
                numba.uint8[:],
                numba.int32[:]),
                cache=True)
def count_node_snps(node_snps,preordered_nodes,parent_nodes,seed_node,reconstructed_alleles,base_pattern_columns):
    # Note that preordered node list does not include the root
    for node_index in preordered_nodes:
        parent_node_index = parent_nodes[node_index]
        if reconstructed_alleles[node_index] < 4 and reconstructed_alleles[parent_node_index] < 4 \
          and reconstructed_alleles[node_index] != reconstructed_alleles[parent_node_index]:
            node_snps[node_index] += len(base_pattern_columns)

# Reconstruct missing data at internal nodes
############################################
@njit(numba.void(numba.uint8[:],
                numba.int32[:],
                numba.int32[:],
                numba.int32[:],
                numba.uint8[:],
                numba.int32[:,:],
                numba.uint8[:]),
                cache=True)
def reconstruct_alleles(reconstructed_alleles,
                        postordered_nodes,
                        leaf_nodes,
                        node_index_to_aln_row,
                        column,
                        child_nodes,
                        reconstructed_base_indices):
    for node_index in postordered_nodes:
        if node_index in leaf_nodes:
            alignment_index = node_index_to_aln_row[node_index]
            reconstructed_alleles[node_index] = column[alignment_index]
        else:
            has_child_base = False
            for child_taxon_index in child_nodes[node_index,:]:
                if child_taxon_index > -1:
                    if reconstructed_alleles[child_taxon_index] < 4:
                        has_child_base = True
            if has_child_base:
                reconstructed_alleles[node_index] = reconstructed_base_indices[node_index]
            else:
                reconstructed_alleles[node_index] = numpy.uint8(4)

# Transfer reconstructed alleles into alignment
###############################################
@njit(numba.void(numba.typeof(numpy.dtype('i1'))[:,:],
                numba.uint8[:],
                numba.typeof(numpy.dtype('i1'))[:],
                numba.int32[:],
                numba.int32[:]),
                cache=True)
def fill_out_aln(out_aln,reconstructed_alleles,ordered_bases,ancestral_node_order,base_pattern_columns):
    for index in numpy.arange(ancestral_node_order.size, dtype=numpy.int32):
        node_index = ancestral_node_order[index]
        base_index = reconstructed_alleles[node_index]
        base = ordered_bases[base_index]
        for column in base_pattern_columns:
            out_aln[column,index] = base

# Reconstruct each base pattern
###############################
@njit(numba.void(numba.uint8[:],
                numba.int32[:],
                numba.float32[:,:],
                numba.uint8[:,:],
                numba.typeof(numpy.dtype('i1'))[:,:],
                numba.typeof(numpy.dtype('i1'))[:],
                numba.int32[:],
                numba.int32[:],
                numba.int32[:],
                numba.int32[:,:],
                numba.int32,
                numba.int32[:],
                numba.int32[:],
                numba.float32[:,:],
                numba.float32[:],
                numba.int32[:],
                numba.uint8[:],
                numba.int32[:]),
                cache=True)
def iterate_over_base_patterns(column,
                                base_pattern_columns,
                                Lmat,
                                Cmat,
                                tmp_out_aln,
                                ordered_bases,
                                postordered_nodes,
                                preordered_nodes,
                                parent_nodes,
                                child_nodes,
                                seed_node,
                                leaf_nodes,
                                ancestral_node_order,
                                node_pij,
                                base_frequencies,
                                node_index_to_aln_row,
                                reconstructed_base_indices,
                                node_snps):

    Cmat_null = numpy.array([0,1,2,3], dtype = numpy.uint8)
        
    # Reset matrices
    Lmat.fill(-numpy.inf)
    Cmat[:] = Cmat_null

    # Count unknown bases
    unknown_base_count = numpy.count_nonzero(column > 3)
    column_base_indices = numpy.unique(column[numpy.where(column <= 3)])
      
    # Heuristic for speed: if all taxa are monomorphic, with a gap in only one sequence, then the ancestral states
    # will all be the observed base, as no ancestral node will have two child nodes with unknown bases at this site
    if unknown_base_count == 1 and column_base_indices.size == 1:
        # If site is monomorphic - replace entire column
        tmp_out_aln[base_pattern_columns,:] = ordered_bases[column_base_indices[0]]
    else:
        # Otherwise perform a full ML inference
        #1 For each OTU y perform the following:
        #Visit a nonroot internal node, z, which has not been visited yet, but both of whose sons, nodes x and y, have already been visited, i.e., Lx(j), Cx(j), Ly(j), and Cy(j) have already been defined for each j. Let tz be the length of the branch connecting node z and its father. For each amino acid i, compute Lz(i) and Cz(i) according to the following formulae:
        #Denote the three sons of the root by x, y, and z. For each amino acid k, compute the expression Pk x Lx(k) x Ly(k) x Lz(k). Reconstruct r by choosing the amino acid k maximizing this expression. The maximum value found is the likelihood of the best reconstruction.
        for node_index in postordered_nodes:
            if node_index == seed_node:
                continue
            #calculate the transistion matrix for the branch
            pij=numpy.reshape(node_pij[node_index,:].copy(),(4,4))
            if node_index in leaf_nodes:
                alignment_index = node_index_to_aln_row[node_index]
                taxon_base_index = column[alignment_index]
                process_leaf(Lmat,
                                Cmat,
                                pij,
                                node_index,
                                taxon_base_index)
            else:
                #2a. Lz(i) = maxj Pij(tz) x Lx(j) x Ly(j)
                #2b. Cz(i) = the value of j attaining the above maximum.
                find_most_likely_base_given_descendents(Lmat,
                                                        Cmat,
                                                        pij,
                                                        node_index,
                                                        child_nodes,
                                                        column_base_indices)

        # Calculate likelihood of base at root node
        child_node_indices = child_nodes[node_index,:]
        child_node_indices = child_node_indices[child_node_indices > -1]
        calculate_root_likelihood(Lmat, Cmat, base_frequencies, node_index, child_node_indices, column_base_indices)
        max_root_base_index = Cmat[node_index,numpy.argmax(Lmat[node_index,:])]
        reconstructed_base_indices[node_index] = max_root_base_index
        
        #Traverse the tree from the root in the direction of the OTUs, assigning to each node its most likely ancestral character as follows:
        # Note that preordered node list does not include the root
        for node_index in preordered_nodes:
            #5a. Visit an unreconstructed internal node x whose father y has already been reconstructed. Denote by i the reconstructed amino acid at node y.
            parent_node_index = parent_nodes[node_index]
            i = reconstructed_base_indices[parent_node_index]
            #5b. Reconstruct node x by choosing Cx(i).
            reconstructed_base_indices[node_index] = Cmat[node_index,i]

        # Put gaps back in and check that any ancestor with only gaps downstream is made a gap
        # store reconstructed alleles
        reconstructed_alleles = numpy.full(postordered_nodes.size, 8, dtype = numpy.uint8)
        reconstruct_alleles(reconstructed_alleles,
                            postordered_nodes,
                            leaf_nodes,
                            node_index_to_aln_row,
                            column,
                            child_nodes,
                            reconstructed_base_indices
                            )

        # If site is not monomorphic - replace specific entries
        fill_out_aln(tmp_out_aln,
                    reconstructed_alleles,
                    ordered_bases,
                    ancestral_node_order,
                    base_pattern_columns
                    )

        # enumerate the number of base subtitutions reconstructed occurring on each branch
        count_node_snps(node_snps,
                        preordered_nodes,
                        parent_nodes,
                        seed_node,
                        reconstructed_alleles,
                        base_pattern_columns,
                        )

# Convert integers to bases
###########################

@njit(numba.void(numba.int8[:],
                 numba.typeof(numpy.dtype('U1'))[:]),
                cache = True)
def int_to_seq(seq,out_seq):

    for i,b in enumerate(seq):
        if b == 0:
            out_seq[i] = 'A'
        elif b == 1:
            out_seq[i] = 'C'
        elif b == 2:
            out_seq[i] = 'G'
        elif b == 3:
            out_seq[i] = 'T'
        elif b == 4:
            out_seq[i] = '-'
        elif b == 5:
            out_seq[i] = 'N'
        else:
            print('Unable to process integer')

##################
# Main functions #
##################

####################################################
# Function for converting alignment to numpy array #
####################################################

def get_base_patterns(prefix, verbose, threads = 1):
    
    # Identify unique base patterns
    if verbose:
        print("Finding unique base patterns")
    t1=time.process_time()
    
    # Check njit function is compiled before multiprocessing
    try:
        seq_to_int()
    except:
        pass

    # Read in ordered sequence names
    sequence_names_fn = prefix + '.gaps.sequence_names.csv'
    sequence_names = []
    if not os.path.isfile(sequence_names_fn):
        sys.exit("Unable to open sequence names file " + sequence_names_fn + "\n")
    with open(sequence_names_fn,'r') as names_file:
        for line in names_file:
            sequence_names.append(line.rstrip())

    # Convert alignment to Numpy array
    codec = 'utf-32-le' if sys.byteorder == 'little' else 'utf-32-be'
    # Read in the unique base patterns
    base_patterns_fn = prefix + '.gaps.base_patterns.csv'
    if not os.path.isfile(base_patterns_fn):
        sys.exit("Unable to open base patterns file " + base_patterns_fn + "\n")
    array_of_pattern_arrays = []
    with open(base_patterns_fn,'r') as pattern_file:
        for line in pattern_file:
            unicode_seq = numpy.frombuffer(bytearray(line.rstrip(), codec), dtype = 'U1')
            out_array = numpy.ndarray(unicode_seq.shape, dtype = numpy.uint8)
            seq_to_int(unicode_seq,out_array)
            array_of_pattern_arrays.append(out_array)
    vstacked_patterns = numpy.vstack(array_of_pattern_arrays)
    # Read in base positions
    array_of_position_arrays = []
    base_positions_fn = prefix + '.gaps.base_positions.csv'
    if not os.path.isfile(base_positions_fn):
        sys.exit("Unable to open base positions file " + base_positions_fn + "\n")
    array_max = 0
    with open(base_positions_fn, 'r') as positions_file:
        for line in positions_file:
            subarray = numpy.asarray(list(map(int,line.rstrip().split(','))),
                                    dtype = numpy.int32)
            array_of_position_arrays.append(subarray)
            subarray_max = numpy.amax(subarray)
            if subarray_max > array_max:
                array_max = subarray_max
    array_max = array_max + 1

    # Record timing
    t2=time.process_time()
    if verbose:
        print("Time taken to load unique base patterns:", t2-t1, "seconds")
        print("Unique base patterns: ", array_max)
    
    # Return output
    return sequence_names,vstacked_patterns,array_of_position_arrays,array_max

########################################################
# Function for reconstructing individual base patterns #
########################################################

def reconstruct_alignment_column(column_index,
                                base_pattern_positions,
                                tree = None,
                                preordered_nodes = None,
                                postordered_nodes = None,
                                leaf_nodes = None,
                                parent_nodes = None,
                                child_nodes = None,
                                seed_node = None,
                                node_pij = None,
                                node_index_to_aln_row = None,
                                ancestral_node_order = None,
                                base_patterns = None,
                                base_frequencies = None,
                                new_aln = None,
                                threads = 1,
                                verbose = False):
    
    ### TIMING
    if verbose:
        prep_time = 0.0
        calc_time = 0.0
        prep_time_start = time.process_time()

    # Load shared memory output alignment
    out_aln_shm = shared_memory.SharedMemory(name = new_aln.name)
    out_aln = numpy.ndarray(new_aln.shape, dtype = numpy.int8, buffer = out_aln_shm.buf)
    
    # Generate data structures for reconstructions
    num_nodes = len(tree.nodes())
    Lmat = numpy.full((num_nodes,4), -numpy.inf, dtype = numpy.float32)
    Cmat = numpy.full((num_nodes,4), [0,1,2,3], dtype = numpy.uint8)
    reconstructed_base_indices = numpy.full(num_nodes, 8, dtype = numpy.uint8)

    # Record SNPs reconstructed as occurring on each branch
    
    ordered_bases = numpy.array([0,1,2,3,4], dtype=numpy.int8)
    node_snps = numpy.zeros(num_nodes, dtype = numpy.int32)

    # Load base pattern information
    base_patterns_shm = shared_memory.SharedMemory(name = base_patterns.name)
    base_patterns = numpy.ndarray(base_patterns.shape, dtype = base_patterns.dtype, buffer = base_patterns_shm.buf)
    # Load base pattern position information
    
    # Extract information for iterations
    column = base_patterns[column_index]

    ### TIMING
    if verbose:
        prep_time_end = time.process_time()
        prep_time = prep_time_end - prep_time_start
        calc_time_start = time.process_time()

    # Iterate over columns
    iterate_over_base_patterns(column,
                                base_pattern_positions,
                                Lmat,
                                Cmat,
                                out_aln,
                                ordered_bases,
                                postordered_nodes,
                                preordered_nodes,
                                parent_nodes,
                                child_nodes,
                                seed_node,
                                leaf_nodes,
                                ancestral_node_order,
                                node_pij,
                                base_frequencies,
                                node_index_to_aln_row,
                                reconstructed_base_indices,
                                node_snps)

    # Close shared memory
    out_aln_shm.close()
    base_patterns_shm.close()

    return node_snps

##################################################
# Function for reconstructing complete alignment #
##################################################

def jar(sequence_names = None,
        base_patterns = None,
        base_pattern_positions = None,
        alignment_filename = None,
        tree_filename = None,
        info_filename = None,
        info_filetype = None,
        output_prefix = None,
        outgroup_name = None,
        threads = 1,
        verbose = False,
        mp_method = "spawn",
        max_pos = None):

    if verbose:
        prep_time = 0.0
        calc_time = 0.0
        prep_time_start = time.process_time()
    # Create a new alignment for the output containing all taxa in the input alignment
    alignment_sequence_names = {}
    for i, name in enumerate(sequence_names):
        alignment_sequence_names[name] = i
    
    # Read the tree
    if verbose:
        print("Reading tree file:", tree_filename)
    tree=read_tree(tree_filename)
    
    # Read the info file and get frequencies and rates
    if info_filename != "":
        if verbose:
            print("Reading info file:", info_filename)
        f,r = read_info(info_filename, type = info_filetype)
    else:
        if verbose:
            print("Using default JC rates and frequencies")
        f=numpy.array([0.25,0.25,0.25,0.25], dtype = numpy.float32)
        r=numpy.array([1.0,1.0,1.0,1.0,1.0,1.0], dtype = numpy.float32)
    
    if verbose:
        print("Frequencies:", ", ".join(map(str,f)))
        print("Rates:", ", ".join(map(str,r)))
    
    # Create rate matrix from f and r
    rm = create_rate_matrix(f,r)

    # Label internal nodes in tree and add these to the new alignment and calculate pij per non-root branch
    nodecounter=0
    num_nodes = len(tree.nodes())
    node_indices = {}
    child_nodes_array = numpy.empty(num_nodes, dtype=object)
    leaf_node_list = []
    node_labels = numpy.empty(num_nodes, dtype=object)
    node_pij = numpy.full((num_nodes,16), -numpy.inf, dtype=numpy.float32)
    postordered_nodes = numpy.arange(num_nodes, dtype=numpy.int32)
    seed_node = None
    seed_node_edge_truncation = True
    node_index_to_aln_row = numpy.full(num_nodes, -1, dtype=numpy.int32)
    ancestral_node_indices = {}

    for node_index,node in zip(postordered_nodes,tree.postorder_node_iter()):
        if node.taxon == None:
            nodecounter+=1
            nodename="Node_"+str(nodecounter)
            tree.taxon_namespace.add_taxon(dendropy.Taxon(nodename))
            node.taxon=tree.taxon_namespace.get_taxon(nodename)
            if nodename in alignment_sequence_names:
                print(nodename, "already in alignment. Quitting")
                sys.exit(209)
            ancestral_node_indices[node_index] = nodename
        else:
            node.taxon.label = node.taxon.label.strip("'")
            if node.taxon.label in alignment_sequence_names:
                node_index_to_aln_row[node_index] = alignment_sequence_names[node.taxon.label]
            else:
                sys.stderr.write('Unable to find ' + node.taxon.label + ' in alignment')
                sys.exit(1)
        if node.parent_node == None:
            seed_node = node_index
        elif node.parent_node == tree.seed_node and seed_node_edge_truncation:
            # Set the length of one root-to-child branch to ~zero
            # as reconstruction should occur with rooting at a node
            # midpoint rooting causes problems at the root, especially w/JC69
            # With outgroup:
            # Prefer to truncate the ingroup branch to force the recombinations
            # onto the outgroup branch
            if outgroup_name is None or \
                            (outgroup_name is not None and node.taxon.label != outgroup_name):
                seed_node_edge_truncation = False
                node_pij[node_index,:]=calculate_pij(node.edge_length/1e6, rm)
            else:
                node_pij[node_index,:]=calculate_pij(node.edge_length, rm)
        else:
            node_pij[node_index,:]=calculate_pij(node.edge_length, rm)
        # Store information to avoid subsequent recalculation as
        # look up of taxon labels with dendropy is slower than native data structures
        node_label = node.taxon.label
        node_indices[node_label] = node_index
        node_labels[node_index] = node_label
        if node.is_leaf():
            leaf_node_list.append(node_index)
            child_nodes_array[node_index] = numpy.full(1, -1, dtype=numpy.int32) # Cannot leave array empty
        else:
            child_nodes_array[node_index] = numpy.array([node_indices[child.taxon.label] for child in node.child_node_iter()],
                                                    dtype=numpy.int32)
    leaf_nodes = numpy.array(leaf_node_list, dtype = numpy.int32)
    child_nodes = convert_to_square_numpy_array(child_nodes_array)

    # Store the preordered nodes and record parent node information
    parent_nodes = numpy.full(num_nodes, -1, dtype = numpy.int32)
    preordered_nodes = numpy.full(num_nodes-1, -1, dtype=numpy.int32)
    for node_count,node in enumerate(tree.preorder_node_iter()):
        if node.parent_node != None:
            node_index = node_indices[node.taxon.label]
            preordered_nodes[node_count-1] = node_index # Do not add root node to preordered nodes
            parent_nodes[node_index] = node_indices[node.parent_node.taxon.label]

        
    ## Switch this to a numpy.int8 data type use the default as 5 the corresponding value to N
    ## form the seq_to_int transformation o the sequence 
    new_aln_array = numpy.full((max_pos,len(ancestral_node_indices)), 5, dtype = numpy.int8)
    
    # Index names for reconstruction
    ancestral_node_order = numpy.fromiter(ancestral_node_indices.keys(), dtype=numpy.int32)

    # Compile functions prior to multiprocessing
    for func in [find_most_likely_base_given_descendents,
                process_leaf,
                calculate_root_likelihood,
                count_node_snps,
                reconstruct_alleles,
                fill_out_aln,
                iterate_over_base_patterns,
                int_to_seq]:
        try:
            func()
        except:
            pass

    # Reconstruct each base position
    if verbose:
        print("Reconstructing sites on tree")

    with SharedMemoryManager() as smm:
    
        # Convert alignment to shared memory numpy array
        new_aln_shared_array = generate_shared_mem_array(new_aln_array, smm)
        
        # Convert base patterns to shared memory numpy array
        base_patterns_shared_array = generate_shared_mem_array(base_patterns, smm)
        
        # Convert base pattern positions to shared memory numpy array
        bp_list = list(range(len(base_patterns)))
        npatterns = len(base_patterns)

        if verbose:
            prep_time_end = time.process_time()
            prep_time = prep_time_end - prep_time_start
            calc_time_start = time.process_time()


        if threads > 1:

            # Parallelise reconstructions across alignment columns using multiprocessing
            with multiprocessing.get_context(method=mp_method).Pool(processes = threads) as pool:
                reconstruction_results = pool.starmap(partial(
                                            reconstruct_alignment_column,
                                                tree = tree,
                                                preordered_nodes = preordered_nodes,
                                                postordered_nodes = postordered_nodes,
                                                leaf_nodes = leaf_nodes,
                                                parent_nodes = parent_nodes,
                                                child_nodes = child_nodes,
                                                seed_node = seed_node,
                                                node_pij = node_pij,
                                                node_index_to_aln_row = node_index_to_aln_row,
                                                ancestral_node_order = ancestral_node_order,
                                                base_patterns = base_patterns_shared_array,
                                                base_frequencies = f,
                                                new_aln = new_aln_shared_array,
                                                threads = threads,
                                                verbose = verbose),
                                            zip(bp_list, base_pattern_positions)
                                        )
            
            # Write out alignment while shared memory manager still active
            out_aln_shm = shared_memory.SharedMemory(name = new_aln_shared_array.name)
            out_aln = numpy.ndarray(new_aln_array.shape, dtype = 'i1', buffer = out_aln_shm.buf)
                     
            # Release pool nodes
            pool.join()
        
        else:
        
            # Run as a loop on a single core
            reconstruction_results = \
                [
                            reconstruct_alignment_column(
                                b,
                                p,
                                tree = tree,
                                preordered_nodes = preordered_nodes,
                                postordered_nodes = postordered_nodes,
                                leaf_nodes = leaf_nodes,
                                parent_nodes = parent_nodes,
                                child_nodes = child_nodes,
                                seed_node = seed_node,
                                node_pij = node_pij,
                                node_index_to_aln_row = node_index_to_aln_row,
                                ancestral_node_order = ancestral_node_order,
                                base_patterns = base_patterns_shared_array,
                                base_frequencies = f,
                                new_aln = new_aln_shared_array,
                                threads = threads,
                                verbose = verbose) for b,p in zip(bp_list, base_pattern_positions)
                ]
            
            # Extract final result
            out_aln_shm = shared_memory.SharedMemory(name = new_aln_shared_array.name)
            out_aln = numpy.ndarray(new_aln_array.shape, dtype = 'i1', buffer = out_aln_shm.buf)

    # Process outputs
    aln_line = numpy.full(len(out_aln[:,0]),"?",dtype="U1")
    if verbose:
        print("Printing alignment with internal node sequences: ", output_prefix+".joint.aln")
    source = alignment_filename
    destination = output_prefix+".joint.aln"
    dest = shutil.copy(source, destination)
    with open(dest, "a") as asr_output:
        for i,node_index in enumerate(ancestral_node_order):
            taxon = ancestral_node_indices[node_index]
            asr_output.write('>' + taxon + '\n')
            int_to_seq(out_aln[:,i], aln_line)
            asr_output.write(''.join(aln_line) + "\n")

    # Combine results for each base across the alignment
    for node in tree.preorder_node_iter():
        node.edge_length = 0.0 # reset lengths to convert to SNPs
        node_index = node_indices[node.taxon.label]
        for x in range(len(reconstruction_results)):
            try:
                node.edge_length += reconstruction_results[x][node_index];
            except AttributeError:
                continue

    ### TIMING
    if verbose:
        calc_time_end = time.process_time()
        calc_time = (calc_time_end - calc_time_start)

    # Print tree
    from gubbins.common import tree_as_string
    
    if verbose:
        print("Printing tree with internal nodes labelled: ", output_prefix+".joint.tre")
    with open(output_prefix+".joint.tre", "w") as tree_output:
    
        recon_tree = tree_as_string(tree,
                                    suppress_rooting=True,
                                    suppress_internal=False)
        print(recon_tree.replace('\'', ''),
              file = tree_output)

    if verbose:
        print("Done")
        print('Time for JAR preparation:\t' + str(prep_time))
        print('Time for JAR calculation:\t' + str(calc_time))




#! python

# encoding: utf-8
# Wellcome Trust Sanger Institute and Imperial College London
# Copyright (C) 2020  Wellcome Trust Sanger Institute and Imperial College London
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

# Generic imports
import sys
import argparse
import re
# Phylogenetic imports
import dendropy
# Biopython imports
from Bio import AlignIO
from Bio import Phylo
from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Seq import Seq

# command line parsing
def get_options():

    parser = argparse.ArgumentParser(description='Extract a clade from a Gubbins output',
                                     prog='extract_gubbins_clade')

    # input options
    parser.add_argument('--list',
                        help = 'List of sequences to extract',
                        required = True)
    parser.add_argument('--aln',
                        help = 'Input alignment (FASTA format)',
                        required = True)
    parser.add_argument('--gff',
                        help = 'GFF of recombinant regions detected by Gubbins',
                        required = True)
    parser.add_argument('--tree',
                        help = 'Final tree generated by Gubbins',
                        required = True)
    parser.add_argument('--out',
                        help = 'Output file prefix',
                        required = True)
    parser.add_argument('--out-fmt',
                        help = 'Format of output alignment',
                        default = 'fasta')
    parser.add_argument('--missing-char',
                        help = 'Character used to replace recombinant sequence',
                        default = '-')

    return parser.parse_args()

# main code
if __name__ == "__main__":

    # Get command line options
    args = get_options()
    
    # Parse list of input sequences
    subset = set()
    # Read in FASTA assemblies
    with open(args.list,'r') as seq_list:
        for line in seq_list.readlines():
            subset.add(line.strip().split()[0])
    
    # Extract from alignment
    output_aln_name = args.out + '.aln'
    names_in_alignment = set()
    with open(output_aln_name,'w') as out_aln:
        alignment = AlignIO.read(args.aln,'fasta')
        for taxon in alignment:
            names_in_alignment.add(taxon.id)
            if taxon.id in subset:
                SeqIO.write(taxon, out_aln, args.out_fmt)

    # Check subset sequences are found in alignment
    not_found_in_dataset = subset - names_in_alignment
    if len(not_found_in_dataset) > 0:
        sys.stderr.write('Sequences in subset missing from alignment: ' + \
                            str(not_found_in_dataset) + '\n')
        sys.exit(1)
    
    # Prune from the tree
    output_tree_name = args.out + '.tree'
    tree = dendropy.Tree.get(path = args.tree,
                             schema = 'newick',
                             preserve_underscores = True)
    tree.retain_taxa_with_labels(subset)
    
    # Output tree
    clade_tree_string = tree.as_string(
                        schema='newick',
                        suppress_leaf_taxon_labels=False,
                        suppress_leaf_node_labels=True,
                        suppress_internal_taxon_labels=False,
                        suppress_internal_node_labels=False,
                        suppress_rooting=True,
                        suppress_edge_lengths=False,
                        unquoted_underscores=True,
                        preserve_spaces=False,
                        store_tree_weights=False,
                        suppress_annotations=True,
                        annotations_as_nhx=False,
                        suppress_item_comments=True,
                        node_label_element_separator=' '
                    )
    with open(output_tree_name,'w') as tree_out:
        tree_out.write(clade_tree_string.replace('\'', '') + '\n')

    # Identify relevant recombination blocks
    output_gff_name = args.out + '.gff'
    taxon_pattern = re.compile('taxa="([^"]*)"')
    with open(args.gff,'r') as in_gff, open(output_gff_name,'w') as out_gff:
        for line in in_gff.readlines():
            if line.startswith('##'):
                out_gff.write(line)
            else:
                info = line.rstrip().split('\t')
                taxon_set = set(taxon_pattern.search(info[8]).group(1).split())
                if not taxon_set.isdisjoint(subset):
                    out_gff.write(line)



#! /usr/bin/env python3
# encoding: utf-8

"""
Integration testing of external dependencies. Likely to be the most brittle tests, but the most important.
"""

import unittest
import os
import sys
import glob
import argparse
import pkg_resources
import shutil
from gubbins import common, run_gubbins

unittest.TestLoader.sortTestMethodsUsing = None
modules_dir = os.path.dirname(os.path.abspath(common.__file__))
data_dir = os.path.join(modules_dir, 'tests', 'data')
working_dir = os.path.join(modules_dir, 'tests')

class TestExternalDependencies(unittest.TestCase):

    # Test pairwise comparisons
    def test_pairwise(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--pairwise",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'pairwise.aln')]))
        exit_code = self.check_for_output_files('pairwise')
        self.cleanup('pairwise')
        assert exit_code == 0

    def test_pairwise_altered_rec_options(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--pairwise",
                                                    "--threads", "1",
                                                    "--p-value", "0.01",
                                                    "--trimming-ratio", "2.0",
                                                    os.path.join(data_dir, 'pairwise.aln')]))
        exit_code = self.check_for_output_files('pairwise')
        self.cleanup('pairwise')
        assert exit_code == 0

    def test_pairwise_altered_recon(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--pairwise",
                                                    "--threads", "1",
                                                    "--mar",
                                                    os.path.join(data_dir, 'pairwise.aln')]))
        exit_code = self.check_for_output_files('pairwise')
        self.cleanup('pairwise')
        assert exit_code == 0

    def test_for_outgroup_present(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "fasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--outgroup", "sequence_10",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test individual tree builders
    def test_fasttree(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "fasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_fasttree_seed(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "fasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--seed","42",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_veryfasttree(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "veryfasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_veryfasttree_seed(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "veryfasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--seed","42",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test resuming a default analysis
    def test_veryfasttree_resume(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "veryfasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--no-cleanup",
                                                    "--prefix", "original",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        common.parse_and_run(parser.parse_args(["--tree-builder", "veryfasttree",
                                            "--verbose", "--iterations", "3",
                                            "--resume", "multiple_recombinations.iteration_1.tre",
                                            "--threads", "1",
                                            os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('original')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_hybrid_option(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "hybrid",
                                                "--best-model",
                                                "--verbose", "--iterations", "3",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_seed(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--seed","42",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_custom_model(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--custom-model", "GTR+R2",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_best_model(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                "--best-model",
                                                "--verbose", "--iterations", "3",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_fast(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree-fast",
                                                "--verbose", "--iterations", "3",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_with_dates(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--date",os.path.join(data_dir, 'taxon.times'),
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_with_dates_for_recon(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--recon-with-dates",
                                                "--date",os.path.join(data_dir, 'taxon.times'),
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_with_dates_and_outgroup(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--date", os.path.join(data_dir, 'taxon.times'),
                                                "--outgroup", "sequence_10",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test resuming a default analysis
    def test_iqtree_resume(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--no-cleanup",
                                                    "--prefix", "original",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                            "--verbose", "--iterations", "3",
                                            "--resume", "multiple_recombinations.iteration_1.tre",
                                            "--threads", "1",
                                            os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('original')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test resuming a default analysis
    def test_iqtree_fast_resume(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree-fast",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--no-cleanup",
                                                    "--prefix", "original",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree-fast",
                                            "--verbose", "--iterations", "3",
                                            "--resume", "multiple_recombinations.iteration_1.tre",
                                            "--threads", "1",
                                            os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('original')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxml(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxml_seed(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--seed","42",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxml_custom_model(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--custom-model","GTRCATI",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test resuming a default analysis
    def test_raxml_resume(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--no-cleanup",
                                                    "--prefix", "original",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                            "--verbose", "--iterations", "3",
                                            "--resume", "multiple_recombinations.iteration_1.tre",
                                            "--threads", "1",
                                            os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('original')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxml_quiet(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxmlng(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxmlng",
                                                    "--model","GTR",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxmlng_seed(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxmlng",
                                                    "--model","GTR",
                                                    "--verbose", "--iterations", "3",
                                                    "--seed","42",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxmlng_custom_model(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxmlng",
                                                    "--model","GTR",
                                                    "--verbose", "--iterations", "3",
                                                    "--custom-model","HKY+G",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test resuming a default analysis
    def test_raxmlng_resume(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxmlng",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--no-cleanup",
                                                    "--prefix", "original",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxmlng",
                                            "--verbose", "--iterations", "3",
                                            "--resume", "multiple_recombinations.iteration_1.tre",
                                            "--threads", "1",
                                            os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('original')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_rapidnj(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "rapidnj",
                                                    "--model","JC",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        # Copy file for subsequent tests
        shutil.copyfile(os.path.join(data_dir, 'multiple_recombinations.recombination_predictions.embl'),
                        os.path.join(data_dir, 'new_rapidnj_jc_output.recombination_predictions.embl'))
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test resuming a default analysis
    def test_rapidnj_resume(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "rapidnj",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--model","JC",
                                                    "--no-cleanup",
                                                    "--prefix", "original",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        common.parse_and_run(parser.parse_args(["--tree-builder", "rapidnj",
                                            "--verbose", "--iterations", "3",
                                            "--resume", "multiple_recombinations.iteration_1.tre",
                                            "--model","JC",
                                            "--threads", "1",
                                            os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('original')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def check_rapidnj_consistency(self):
        new_file = os.path.join(data_dir,'new_rapidnj_jc_output.recombination_predictions.embl')
        reference_file = os.path.join(data_dir,'ref_rapidnj_jc_output.recombination_predictions.embl')
        assert common.have_recombinations_been_seen_before(reference_file,[new_file])

    def test_defined_starting_tree(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--starting-tree",
                                                os.path.join(data_dir, 'starting_tree.tre'),
                                                "--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_defined_mislabelled_starting_tree(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--starting-tree",
                                                os.path.join(data_dir, 'mislabelled.starting_tree.tre'),
                                                "--tree-builder", "iqtree",
                                                "--verbose", "--iterations", "3",
                                                "--threads", "1",
                                                os.path.join(data_dir, 'mislabelled.multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('mislabelled.multiple_recombinations')
        self.cleanup('mislabelled.multiple_recombinations')
        assert exit_code == 0

    # Test initial star tree
    def test_starting_star_fasttree(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--first-tree-builder","star",
                                                    "--tree-builder", "fasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_starting_star_iqtree(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--first-tree-builder","star",
                                                    "--tree-builder", "iqtree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_starting_star_iqtree_fast(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--first-tree-builder","star",
                                                    "--tree-builder", "iqtree-fast",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_starting_star_raxml(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--first-tree-builder","star",
                                                    "--tree-builder", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_starting_star_raxmlng(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--first-tree-builder","star",
                                                    "--tree-builder", "raxmlng",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_starting_star_rapidnj(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--first-tree-builder","star",
                                                    "--tree-builder", "rapidnj",
                                                    "--model","JC",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test sequence reconstruction
    def test_raxml_seq_recon(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--seq-recon", "raxml",
                                                    "--mar",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_seq_recon(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--seq-recon", "iqtree",
                                                    "--mar",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxmlng_seq_recon(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--seq-recon", "raxmlng",
                                                    "--mar",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test model fitting
    def test_fasttree_model_fit(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--model-fitter", "fasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_iqtree_model_fit(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--model-fitter", "iqtree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxml_model_fit(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--model-fitter", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_raxmlng_model_fit(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--model-fitter", "raxmlng",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test different model structures across iterations
    def test_different_model_selections(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--model-fitter", "iqtree",
                                                    "--first-model", "JC",
                                                    "--model", "GTRGAMMA",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test bootstrapping
    def test_raxml_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--model","HKY",
                                                    "--bootstrap","10",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_iqtree_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                    "--verbose", "--iterations", "3",
                                                    "--model","HKY",
                                                    "--bootstrap","1000",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_rapidnj_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "rapidnj",
                                                    "--model","JC",
                                                    "--verbose", "--iterations", "3",
                                                    "--bootstrap","10",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_raxmlng_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxmlng",
                                                    "--model","HKY",
                                                    "--verbose", "--iterations", "3",
                                                    "--bootstrap","100",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_fasttree_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "fasttree",
                                                    "--verbose", "--iterations", "3",
                                                    "--bootstrap","10",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_fasttree_bootstrapping_with_raxml(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "fasttree",
                                                    "--model","JC",
                                                    "--model-fitter", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--bootstrap","10",
                                                    "--threads", "1",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_raxmlng_transfer_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxmlng",
                                                    "--model","HKY",
                                                    "--verbose", "--iterations", "3",
                                                    "--bootstrap","100",
                                                    "--threads", "1",
                                                    "--transfer-bootstrap",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_raxml_transfer_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--model","HKY",
                                                    "--verbose", "--iterations", "3",
                                                    "--bootstrap","10",
                                                    "--threads", "1",
                                                    "--transfer-bootstrap",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_iqtree_transfer_bootstrapping(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                    "--model","HKY",
                                                    "--verbose", "--iterations", "3",
                                                    "--bootstrap","1000",
                                                    "--threads", "1",
                                                    "--transfer-bootstrap",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    # Test SH tests
    def test_iqtree_sh_test(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                    "--verbose", "--iterations", "3",
                                                    "--model","HKY",
                                                    "--threads", "1",
                                                    "--sh-test",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    def test_raxml_sh_test(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--verbose", "--iterations", "3",
                                                    "--model","HKY",
                                                    "--threads", "1",
                                                    "--sh-test",
                                                    os.path.join(data_dir, 'bootstrapping_test.aln')]))
        exit_code = self.check_for_output_files('bootstrapping_test')
        self.cleanup('bootstrapping_test')
        assert exit_code == 0

    # Test combining multiple methods
    def test_method_combination(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "raxml",
                                                    "--best-model",
                                                    "--iterations", "3",
                                                    "--mar",
                                                    "--seq-recon", "raxmlng",
                                                    "--first-tree-builder", "fasttree",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test convergence
    def test_converge_on_rec(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--converge-method", "recombination",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    def test_converge_on_unweighted_rf(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree",
                                                    "--verbose", "--iterations", "3",
                                                    "--threads", "1",
                                                    "--converge-method", "robinson_foulds",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('multiple_recombinations')
        self.cleanup('multiple_recombinations')
        assert exit_code == 0

    # Test renaming of final output
    def test_rename_final_output(self):
        exit_code = 1
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--prefix", "different_prefix",
                                                    "--verbose", "--iterations", "3",
                                                    os.path.join(data_dir, 'multiple_recombinations.aln')]))
        exit_code = self.check_for_output_files('different_prefix')
        self.cleanup('different_prefix')
        assert exit_code == 0

    def test_cleanup(self):
        parser = run_gubbins.parse_input_args()
        common.parse_and_run(parser.parse_args(["--tree-builder", "iqtree", "--verbose", os.path.join(data_dir, 'multiple_recombinations.aln')]))
        self.cleanup('multiple_recombinations')
        assert not glob.glob('multiple_recombinations.aln.*')
        assert not glob.glob('multiple_recombinations.iteration*')

    @staticmethod
    def check_for_output_files(prefix):
        assert os.path.exists(prefix + '.summary_of_snp_distribution.vcf')
        assert os.path.exists(prefix + '.recombination_predictions.embl')
        assert os.path.exists(prefix + '.per_branch_statistics.csv')
        assert os.path.exists(prefix + '.filtered_polymorphic_sites.fasta')
        assert os.path.exists(prefix + '.filtered_polymorphic_sites.phylip')
        assert os.path.exists(prefix + '.recombination_predictions.gff')
        assert os.path.exists(prefix + '.branch_base_reconstruction.embl')
        assert os.path.exists(prefix + '.final_tree.tre')
        assert os.path.exists(prefix + '.node_labelled.final_tree.tre')
        return 0

    @staticmethod
    def cleanup(prefix):
        #os.chdir(working_dir)
        regex_to_remove = prefix + ".*"
        for file in glob.glob(regex_to_remove):
            os.remove(file)
        for dir in glob.glob("./tmp*"):
            if os.path.isdir(dir):
                shutil.rmtree(dir)

if __name__ == "__main__":
    unittest.main(buffer=True)
